; linked array list
;
; Stan Hebben
;
; released under the MIT license (http://www.opensource.org/licenses/mit-license.php)
;
; NOTE: never add elements while iterating
; NOTE: don't run two iterators at a time if one of them modifies the list (removes the element)
; NOTE: the order of elements is NOT preserved (removes will change order)

include fsalloc.asm

LANODESIZE	equ	16	; change for different node sizes

LALIST struct
	
	head	dword	?
	tail	dword	?
	
	lastspace	dword	?
	_size	dword	?
	
LALIST ends

LANODE struct
	
	prev			dword	?
	next			dword	?
	
	elements	dword	LANODESIZE dup(?)
	
LANODE ends

LAITER struct
	
	_list			dword	?
	nextnode	dword	?
	counter		dword	?
	counter2	dword	?
	
LAITER ends

.code
laCreate proc
	
	push esi
	push edi
	
	push sizeof LALIST
	push LMEM_FIXED
	call LocalAlloc
	mov edi, eax
	
	fsAlloc %(sizeof LANODE)
	
	mov (LALIST ptr [edi]).head, eax
	mov (LALIST ptr [edi]).tail, eax
	mov (LANODE ptr [eax]).next, 0
	mov (LANODE ptr [eax]).prev, 0
	mov (LALIST ptr [edi]).lastspace, LANODESIZE
	mov (LALIST ptr [edi])._size, 0
	
	mov eax, edi
	pop edi
	pop esi
	ret
	
laCreate endp

laAdd proc list:dword, value:dword
	
	push esi
	push edi
	
	mov esi, list
	
	mov edx, (LALIST ptr [esi]).lastspace
	add (LALIST ptr [edi])._size, 1
	
	sub edx, 1
	jc _newspace
	
	mov (LALIST ptr [esi]).lastspace, edx
	mov edi, (LALIST ptr [esi]).tail
	mov eax, value
	mov (LANODE ptr [edi]).elements[4*edx], eax
	
	pop edi
	pop esi
	ret
	
_newspace:
	fsAlloc %(sizeof LANODE)
	
	mov edi, eax
	mov eax, (LALIST ptr [esi]).tail
	mov (LANODE ptr [edi]).prev, eax
	mov (LANODE ptr [eax]).next, edi
	mov (LALIST ptr [esi]).lastspace, LANODESIZE - 1
	mov eax, value
	mov (LANODE ptr [edi]).elements[4*(LANODESIZE-1)], eax
	
	mov (LALIST ptr [esi]).tail, edi
	
	pop edi
	pop esi
	ret
	
laAdd endp

; returns a structure to iterate with
laIterateStart proc list:dword
	
	push esi
	push edi
	
	push sizeof LAITER
	push LMEM_FIXED
	call LocalAlloc
	mov edi, eax
	
	mov esi, list
	
	mov (LAITER ptr [edi])._list, esi
	mov eax, (LALIST ptr [esi]).head
	mov (LAITER ptr [edi]).nextnode, eax
	mov eax, (LALIST ptr [esi])._size
	mov (LAITER ptr [edi]).counter, eax
	mov (LAITER ptr [edi]).counter2, LANODESIZE
	
	mov eax, edi
	pop edi
	pop esi
	ret
	
laIterateStart endp

laIterateNext proc iter:dword
	
	push esi
	push edi
	
	mov esi, iter
	cmp (LAITER ptr [esi]).counter, 0
	jz _done
	
	sub (LAITER ptr [esi]).counter, 1
	
	mov edx, (LAITER ptr [esi]).counter2
	sub edx, 1
	jc _nextnode
	
	mov (LAITER ptr [esi]).counter2, edx
	
	mov eax, (LAITER ptr [esi]).nextnode
	mov eax, (LANODE ptr [eax]).elements[4*edx]
	pop edi
	pop esi
	ret
	
_nextnode:
	mov (LAITER ptr [esi]).counter2, LANODESIZE-1
	mov eax, (LAITER ptr [esi]).nextnode
	mov eax, (LANODE ptr [eax]).next
	mov (LAITER ptr [esi]).nextnode, eax
	
	mov eax, (LANODE ptr [eax]).elements[4*(LANODESIZE-1)]
	pop edi
	pop esi
	ret
	
_done:
	mov eax, 0
	pop edi
	pop esi
	ret
	
laIterateNext endp

; remove the element returned by the last laIterateNext call
laIterateRemove proc iter:dword
	
	push esi
	push edi
	
	mov esi, iter
	
	mov edi, (LAITER ptr [esi])._list
	
	sub (LALIST ptr [edi])._size, 1
	mov eax, (LALIST ptr [edi]).tail
	mov edx, (LALIST ptr [edi]).lastspace
	mov eax, (LANODE ptr [eax]).elements[4*edx]
	add (LALIST ptr [edi]).lastspace, 1
	; last element in eax
	
	mov edi, (LAITER ptr [esi]).nextnode
	mov edx, (LAITER ptr [esi]).counter2
	mov (LANODE ptr [edi]).elements[4*edx], eax
	add edx, 1
	mov (LAITER ptr [esi]).counter2, edx
	
	mov eax, (LAITER ptr [esi])._list
	cmp (LALIST ptr [eax]).lastspace, LANODESIZE
	jz _deletenode
	
	pop edi
	pop esi
	ret
	
_deletenode:
	mov esi, (LAITER ptr [esi])._list
	mov eax, (LALIST ptr [esi]).tail
	mov edx, (LANODE ptr [eax]).prev
	
	mov (LALIST ptr [esi]).tail, edx
	mov (LALIST ptr [esi]).lastspace, 0
	fsFree eax, %(sizeof LANODE)
	
	pop edi
	pop esi
	ret
	
laIterateRemove endp

laIterateEnd proc iter:dword
	
	push iter
	call LocalFree
	
	ret
	
laIterateEnd endp

laDestroy proc list:dword
	
	push esi
	push edi
	
	mov esi, list
	
	mov edx, (LLIST ptr [esi]).head
@@:
	mov edi, (LANODE ptr [edx]).next
	fsFree edx, %(sizeof LANODE)
	mov edx, edi
	test edi, edi
	jnz @B
	
	push esi
	call LocalFree
	
	pop edi
	pop esi
	ret
	
laDestroy endp
